"""
CONSTANT Utilities
"""

# Hardware Related Constants
BCM = 'BCM'
BOARD = 'BOARD'
RISING = 'RISING'
FALLING = 'FALLING'
BOTH = 'BOTH'
INPUT = 'INPUT'
OUTPUT = 'OUTPUT'
PULL_UP = 'PULL_UP'
PULL_DOWN = 'PULL_DOWN'
HIGH = 'HIGH'
LOW = 'LOW'

# Color Related Constants
RED = 'RED'
GREEN = 'GREEN'
BLUE = 'BLUE'
YELLOW = 'YELLOW'
MAGENTA = 'MAGENTA'
CYAN = 'CYAN'
WHITE = 'WHITE'

# Motor Direction Related Constants
CLOCKWISE = 'CW'
COUNTER_CLOCKWISE = 'CCW'
